'''
    Code borrowed from https://github.com/lucidrains/DALLE2-pytorch/tree/main/dalle2_pytorch
'''

import torch
import torch.nn as nn
import numpy as np
from .utils import exists
from copy import deepcopy

def cosine_beta_schedule(timesteps, s = 0.008):
    """
    cosine schedule
    as proposed in https://openreview.net/forum?id=-NEXDKk8gZ
    """
    steps = timesteps + 1
    x = torch.linspace(0, timesteps, steps, dtype = torch.float64)
    alphas_cumprod = torch.cos(((x / timesteps) + s) / (1 + s) * torch.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0].item()
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return torch.clip(betas, 0, 0.999)


def linear_beta_schedule(timesteps):
    scale = 1000 / timesteps
    beta_start = scale * 0.0001
    beta_end = scale * 0.02
    return torch.linspace(beta_start, beta_end, timesteps, dtype = torch.float64)


def quadratic_beta_schedule(timesteps):
    scale = 1000 / timesteps
    beta_start = scale * 0.0001
    beta_end = scale * 0.02
    return torch.linspace(beta_start**0.5, beta_end**0.5, timesteps, dtype = torch.float64) ** 2


def sigmoid_beta_schedule(timesteps):
    scale = 1000 / timesteps
    beta_start = scale * 0.0001
    beta_end = scale * 0.02
    betas = torch.linspace(-6, 6, timesteps, dtype = torch.float64)
    return torch.sigmoid(betas) * (beta_end - beta_start) + beta_start

class Diffusion(object):
    def __init__(self, config):
        self.steps = config.steps
        try:
            self.betas = torch.tensor(np.load(config.betas))
            assert(len(self.betas) == self.steps)
            
        except Exception as e:
            if config.sample_strategy == 'linear':
                self.betas = linear_beta_schedule(self.steps)
            elif config.sample_strategy == 'cos':
                self.betas = cosine_beta_schedule(self.steps)
            elif config.sample_strategy == 'quadratic':
                self.betas = quadratic_beta_schedule(self.steps)
            elif config.sample_strategy == 'sigmoid':
                self.betas = sigmoid_beta_schedule(self.steps)
            else:
                raise NotImplementedError
        
        self.alphas = 1.0 - self.betas
        self.alpha_mul = torch.cumprod(self.alphas, dim = 0)
        np.save(config.beta_path, self.betas.numpy())
        
    def add_noise(self, x, device, given_tim = None, given_prob = None):
        '''
            given_prob can be specified to sample time steps at a given probability.
        '''
        self = self.to(device).to(torch.float32)
        x = x.to(device).to(torch.float32)
        
        bsz = x.shape[0]
        ori_shape = x.shape
        x = x.reshape(bsz, -1)

        if not exists(given_prob):
            tim = torch.randint(0, self.steps, (bsz, )).to(device)
        else:
            given_prob = torch.softmax(given_prob, dim = 0, dtype = torch.float32)
            tim = np.random.choice(self.steps, size = bsz, 
                               replace = True, p = given_prob)
            tim = torch.tensor(tim).to(device)
            
        if exists(given_tim):
            tim = torch.tensor(given_tim).to(device)
            
        alpha_muls = torch.gather(self.alpha_mul, dim = 0, index = tim).to(device).unsqueeze(1)
        noise = torch.randn_like(x, dtype = torch.float32, device = device)
        noised_x = torch.sqrt(alpha_muls) * x + torch.sqrt(1 - alpha_muls) * noise
        
        noised_x = noised_x.reshape(*ori_shape)
        noise = noise.reshape(*ori_shape)
        return noised_x.detach().to(torch.float), noise.detach().to(torch.float), tim.detach()

    def to(self, device):
        self.alphas = self.alphas.to(device)
        self.betas = self.betas.to(device)
        self.alpha_mul = self.alpha_mul.to(device)
        return self

    def denoise(self, x, step, pred_noise, device):
        x = x.to(device).to(torch.float32)
        pred_noise = pred_noise.to(device).to(torch.float32)
        self = self.to(device).to(torch.float32)
        
        x_now = (x - self.betas[step] * pred_noise / (torch.sqrt(1 - self.alpha_mul[step]))) * torch.sqrt(1 / self.alphas[step]) 
        if step > 0:
            z = torch.randn_like(x, dtype = torch.float32, device = device)
            x_now += torch.sqrt((1 - self.alpha_mul[step - 1]) / (1 - self.alpha_mul[step]) * self.betas[step]) * z
        return x_now.to(torch.float)
        


import torch
import torch.nn as nn
from copy import deepcopy

class EMA(object):
    def __init__(self, alpha = 0.99, step_range = 1):
        self.alpha = alpha
        self.step_range = step_range
        self.total_steps = 0
        self.shadow = {}
        self.ori = {}
    
    def register(self, model):
        for name, para in model.named_parameters():
            self.shadow[name] = para.cpu().data.clone()
        
    def resume(self, ckpt):
        self.total_steps = ckpt['ema_total_steps']
        for k in ckpt['ema_shadow']:
            self.shadow[k] = ckpt['ema_shadow'][k].cpu()
        
    def update(self, model):
        self.total_steps += 1
        if self.total_steps % self.step_range != 0:
            return
        for name, para in model.named_parameters():
            if name not in self.shadow:
                raise ValueError("unexpected model parameter!")
            self.shadow[name] = self.alpha * self.shadow[name] + (1 - self.alpha) * para.cpu().data.clone()
        return model
    
    def load_model(self, model):
        self.ori = {}
        for name, para in model.named_parameters():
            if name not in self.shadow:
                raise ValueError("unexpected model parameter!")
            self.ori[name] = para.cpu().data.clone()
            para.data = self.shadow[name].to(para.data.device)
        return model

    def restore_model(self, model):
        for name, para in model.named_parameters():
            if name not in self.shadow or name not in self.ori:
                raise ValueError("unexpected model parameter or not stored!")
            para.data = self.ori[name].to(para.data.device)
        return model
    
    def store_ema(self, ckpt):
        ckpt['ema_total_steps'] = self.total_steps
        ckpt['ema_shadow'] = self.shadow
        return ckpt


import torch
import torch.nn as nn

class UpSample(nn.Module):
    def __init__(self, channels, in_size, out_size, 
                 upsample_type = 'conv_trans', upsample_mode = 'nearest'):
        super().__init__()
        if upsample_type == 'conv_trans' and out_size == 2 * in_size:
            self.upsample = nn.ConvTranspose2d(channels, 
                                                channels,
                                                kernel_size = (3, 3),
                                                stride = 2,
                                                padding = 1,
                                                output_padding = 1)
        else:
            self.upsample = nn.Upsample((out_size, out_size), mode = upsample_mode)
        
    def forward(self, x):
        return self.upsample(x)
    
class DownSample(nn.Module):
    def __init__(self, channels, in_size, out_size, 
                 downsample_type = 'conv', downsample_mode = 'avg'):
        super().__init__()
        if downsample_type == 'conv' and in_size == 2 * out_size:
            self.downsample = nn.Conv2d(channels, 
                                    channels,
                                    kernel_size = (3, 3),
                                    stride = 2,
                                    padding = 1)
        else:
            if downsample_mode == 'avg':
                self.downsample = nn.AdaptiveAvgPool2d((out_size, out_size))
            else:
                self.downsample = nn.AdaptiveMaxPool2d((out_size, out_size))
        
    def forward(self, x):
        return self.downsample(x)


from torch.optim import SGD, Adam, AdamW, Adagrad
from .utils import exists

class Optimizer(object):
    def __init__(self, parameters, lr, config):
        '''
            supported optimizer types: sgd, Adam, AdamW
        '''
        if config.optimizer_type == 'sgd':
            self.optimizer = SGD(parameters, 
                                 lr = lr,
                                 momentum = config.momentum,
                                 weight_decay = config.weight_decay)
        elif config.optimizer_type == 'Adam':
            self.optimizer = Adam(parameters, 
                                  lr = lr,
                                  betas = config.adam_betas,
                                  weight_decay = config.weight_decay)
        elif config.optimizer_type == 'AdamW':
            self.optimizer = AdamW(parameters,
                                   lr = lr,
                                   betas = config.adamw_betas,
                                   weight_decay = config.weight_decay)
        else:
            raise NotImplementedError
        
    def step(self):
        self.optimizer.step()
            
    def zero_grad(self):
        self.optimizer.zero_grad()
        
    def state_dict(self):
        return self.optimizer.state_dict()
    
    def load_state_dict(self, state_dict):
        self.optimizer.load_state_dict(state_dict)


import torch
import torch.nn as nn
from .FC import FC
import math
from copy import deepcopy

class SinPositionalEmbedding(nn.Module):
    def __init__(self, embedding_dim, train = True):
        super().__init__()
        self.dim = embedding_dim
        if train:
            self.mlp = nn.Sequential(
                FC(self.dim, self.dim, dropout = 0, activation = 'silu'),
                nn.Linear(self.dim, self.dim)
            )    
        else:
            self.mlp = nn.Identity()
    
    def forward(self, x):
        x = x.reshape(-1, 1)
        dtype, device = x.dtype, x.device
        half_dim = self.dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device = device, dtype = dtype) * -emb)
        emb = x.unsqueeze(1) * emb.unsqueeze(0)
        emb = torch.cat((emb.sin(), emb.cos()), dim = -1).type(dtype)
        return self.mlp(emb)


import torch.nn as nn
import torch
from .utils import get_mask
from .FC import FC

class MultiHeadAttention(nn.Module):
    def __init__(self, input_dim, num_heads = 4):
        nn.Module.__init__(self)
        self.input_dim = input_dim
        self.heads = num_heads
        self.WQ = nn.Linear(input_dim, num_heads * input_dim)
        self.WK = nn.Linear(input_dim, num_heads * input_dim)
        self.WV = nn.Linear(input_dim, num_heads * input_dim)
        self.W = nn.Linear(input_dim * num_heads, input_dim)
        self.sfm = nn.Softmax(dim = -1)
        
    def split_heads(self, Q, K, V, batch_size):
        rQ = Q.view(batch_size, -1, self.heads, self.input_dim).transpose(1, 2)
        rK = K.view(batch_size, -1, self.heads, self.input_dim).transpose(1, 2)
        rV = V.view(batch_size, -1, self.heads, self.input_dim).transpose(1, 2)
        return rQ, rK, rV
    
    def forward(self, Q, K, V, device, mask = None):
        '''
            Q, K, V: [bs, seq_len, dim] -> [bs, seq_len, dim*heads]
            -> [bs, seq_len, heads, dim] -> [bs, heads, seq_len, dim]
            matmul-> [bs, heads, seq_len, seq_len]
        '''
        sQ = self.WQ(Q)
        sK = self.WK(K)
        sV = self.WV(V)
        batch_size = Q.shape[0]

        rQ, rK, rV = self.split_heads(sQ, sK, sV, batch_size)
        Si = torch.matmul(rQ, rK.transpose(2, 3))

        Si = Si / torch.sqrt(torch.tensor([self.input_dim]).to(device))  
        if mask != None:
            Si = Si + mask
            
        sim = self.sfm(Si).to(rV.dtype)
        oV = torch.matmul(sim, rV).transpose(1, 2).reshape(batch_size, -1, self.heads * self.input_dim)
        
        output = self.W(oV)
        return output


class MultiHeadSparseAttention(nn.Module):
    def __init__(self, input_dim, split_length, num_heads):
        nn.Module.__init__(self)
        self.input_dim = input_dim
        self.heads = num_heads
        self.split_length = split_length
        self.WQ = nn.Linear(input_dim, num_heads * input_dim)
        self.WK = nn.Linear(input_dim, num_heads * input_dim)
        self.WV = nn.Linear(input_dim, num_heads * input_dim)
        self.W = nn.Linear(input_dim * num_heads, input_dim)
        self.att = MultiHeadAttention(input_dim, num_heads)
        self.sfm = nn.Softmax(dim = -1)

    def split_heads(self, Q, K, V, batch_size):
        rQ = Q.view(batch_size, Q.shape[1] // self.split_length, self.split_length,
                    self.heads, self.input_dim).transpose(2, 3)
        rK = K.view(batch_size, K.shape[1] // self.split_length, self.split_length, 
                    self.heads, self.input_dim).transpose(2, 3)
        rV = V.view(batch_size, V.shape[1] // self.split_length, self.split_length, 
                    self.heads, self.input_dim).transpose(2, 3)
        return rQ, rK, rV
    
    def forward(self, Q, K, V, device):
        '''
            Q, K, V: [bs, seq_len, dim] -> [bs, seq_len, dim*heads]
            -> [bs, seq_len, heads, dim] -> [bs, heads, seq_len, dim]
            matmul-> [bs, heads, seq_len, seq_len]
        '''
        assert(Q.shape[1] % self.split_length == 0)
        assert(K.shape[1] % self.split_length == 0)
        assert(V.shape[1] % self.split_length == 0)
        sQ = self.WQ(Q)
        sK = self.WK(K)
        sV = self.WV(V)
        bsz = Q.shape[0]
        seq_len = Q.shape[1]

        rQ, rK, rV = self.split_heads(sQ, sK, sV, bsz) #[bsz, split_time, head, split_len, dim]
        Si = torch.matmul(rQ, rK.transpose(3, 4))

        s_mask = get_mask(Si.shape[3])
        Si = Si / torch.sqrt(torch.tensor([self.input_dim]).to(device))
        
        sim = self.sfm(Si + s_mask)  #[bsz, split_time, head, split_len, split_len]
        sim = sim.to(rV.dtype)
        oV = torch.matmul(sim, rV).reshape(bsz, -1, self.heads, 
                                        self.split_length, self.input_dim) 
        #[bsz, split_time, heads, split_len, dim]
        
        pooling = oV.mean(dim = -2).transpose(1, 2) #average-pooling, [bsz, heads, split_tim, dim]
        g_mask = g_mask = get_mask(pooling.shape[2]).to(device)
        g_Si = torch.matmul(pooling, pooling.transpose(2, 3)) / \
                            torch.sqrt(torch.tensor([self.input_dim]).to(device))
                            
        g_sim = self.sfm(g_Si + g_mask).to(pooling.dtype) #[bsz, heads, split_tim, split_tim]
        global_att = torch.matmul(g_sim, pooling).transpose(1, 2) #[bsz, split_tim, heads, dim]
        oV = oV + global_att.unsqueeze(3)
        output = self.W(oV) #[bsz, split_time, split_len, dim]
        output = output.reshape(bsz, -1, self.input_dim)
        return output
    
class DecBlock(nn.Module):
    def __init__(self, dropout, input_dim, num_heads, att_type = '', split_length = 16):
        nn.Module.__init__(self)
        if att_type == 'sparse':
            self.self_att = MultiHeadSparseAttention(input_dim, split_length, num_heads)
        else:
            self.self_att = MultiHeadAttention(input_dim, num_heads)
        self.dropout = nn.Dropout(dropout)
        self.ln2 = nn.LayerNorm(input_dim)
        self.fc = nn.Sequential(
            FC(input_dim, input_dim, dropout, activation = 'gelu'),
            nn.Linear(input_dim, input_dim),
        )
        self.ln3 = nn.LayerNorm(input_dim)

    def forward(self, dec_input, device, mask = None):
        self_att_out = self.self_att(dec_input, dec_input, dec_input, device, mask)
        ln_out2 = self.ln2(self_att_out + dec_input)
        output = self.ln3(self.fc(self.dropout(ln_out2)) + ln_out2)
        return output


from torch.optim.lr_scheduler import CosineAnnealingLR, StepLR, ConstantLR
from .utils import exists

class Scheduler(object):
    def __init__(self, optimizer, config):
        '''
            supported scheduler types: cos_lr, step_lr, const_lr
        '''
        if config.scheduler_type == 'cos_lr':
            self.scheduler = CosineAnnealingLR(optimizer, T_max = config.T_max, eta_min = config.eta_min)
        elif config.scheduler_type == 'step_lr':
            self.scheduler = StepLR(optimizer, step_size = config.step_size, gamma = config.gamma)
        elif config.scheduler_type == 'const_lr':
            self.scheduler = ConstantLR(optimizer, factor = config.factor, total_iters = config.total_iters)
        else:
            self.scheduler = None
        
    def step(self):
        if exists(self.scheduler):
            self.scheduler.step()


import torch
import torch.nn as nn
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data import DataLoader, DistributedSampler
import os

class Parallel_Trainer(object):
    def __init__(self, synch, config):
        self.synch = synch
        if synch:
            local_rank = int(os.environ['LOCAL_RANK'])
            torch.cuda.set_device(local_rank)
            self.device = torch.device("cuda", local_rank)
            self.local_rank = local_rank
            torch.distributed.init_process_group("nccl", world_size = config.n_gpus, 
                                                rank = local_rank, init_method = 'env://')
        else:
            self.device = config.device
    
    def get_device(self):
        '''
            Get your current process's corresponding device, which is useful 
            for both single and multiple GPUs training.
        '''
        return self.device

    def model_parallel(self, model):
        '''
            Set the model to train mode, put it on the corresponding device and 
            initialize it to DDP if synch is set to True.
        '''
        model = model.to(self.device)
        model.train()
        if self.synch:
            model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)   
            model = DistributedDataParallel(model.cuda(self.local_rank), device_ids=[self.local_rank],
                                            find_unused_parameters=True)
            return model
        else:
            return model
    
    def dataset_parallel(self, dataset):
        '''
            Get a distributed dataset sampler for DDP.
        '''
        if self.synch:
            train_sampler = DistributedSampler(dataset)
            return train_sampler
        else:
            return None
    
    def get_loader(self, dataset, epoch, bsz, num_workers, train_sampler = None):
        '''
            Get dataloader for both DDP or non-DDP. 
        '''
        if self.synch:
            train_sampler.set_epoch(epoch)
            dataloader = DataLoader(dataset, bsz, pin_memory=True,
                                num_workers= num_workers, sampler=train_sampler)
            return dataloader
        else:
            dataloader = DataLoader(dataset, bsz, num_workers = num_workers, shuffle = True,
                                    drop_last = False)
        return dataloader
    
    def save_model(self, ckpt, model, path):
        ''''
            Put the model's state_dict to ckpt and save the ckpt at path.
            If you'd like to store other information, you can put them into 
            ckpt outside this function. Only process running on the first GPU
            needs to actually save the model.
        '''
        if self.synch:
            if self.local_rank == 0:
                ckpt['state_dict'] = model.module.state_dict()
                torch.save(ckpt, path)
        
        else:
            ckpt['state_dict'] = model.state_dict()
            torch.save(ckpt, path)
        
    def write_log(self, log_file, content):
        '''
            Similar to saving checkpoints, write content to log_file. Only
            process running on the first GPU really does the work this function.
        '''
        if self.synch:
            if self.local_rank == 0:
                log_file.write(content + '\n')
                log_file.flush()
            
        else:
            log_file.write(content + '\n')
            log_file.flush()
    
    def call_func(self, func, *args, **kwargs):
        '''
            This is a general function caller that assures only the process running 
            on the first GPU actually calls the func.
        '''
        if self.synch:
            if self.local_rank == 0:
                func(*args, **kwargs)
        
        else:
            func(*args, **kwargs)
        
    def get_model(self, model):
        '''
            Returns the actual model. In single GPU training, it's model; in DDP,
            it's model.module.
        '''
        if self.synch:
            return model.module

        else:
            return model


import torch
import os, shutil
import torch.nn as nn
import numpy as np
from torchvision.utils import save_image
from PIL import Image, ImageFile
import math, time, random
ImageFile.LOAD_TRUNCATED_IMAGES = True

def clear_cuda():
    '''
        This function clears out cuda for more spaces. It is extremely useful when you encounter
        an CUDA out-of-memory in the middle of your training process.
    '''
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    time.sleep(5)
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    torch.cuda.empty_cache()
    
def fea2cha(x):
    '''
        This function accepts an input with shape [bsz, H*H, C] and converts it into
        [bsz, C, H, H]. This often occurs in attentions applied on images.
    '''
    assert(len(x.shape) == 3) # Your input should have exactly 3 dimensions.

    H = int(math.sqrt(x.shape[1]))
    C = x.shape[2]
    y = x.transpose(1, 2).reshape(-1, C, H, H)
    return y

def cha2fea(x):
    '''
        This function accepts an input with shape [bsz, C, H, W] and converts it into
        [bsz, H*W, C]. This often occurs in attentions applied on images.
    '''
    assert(len(x.shape) == 4) # Your input should have exactly 4 dimensions.
    
    C = x.shape[1]
    H = x.shape[2]
    W = x.shape[3]
    
    y = x.transpose(1, 2).transpose(2, 3).reshape(-1, H * W, C)
    return y 

def select_activation(activation):
    '''
        This is a simple function used for selecting activations, note that all the parameters are not tuneable.
        activation : 'relu', 'leakyrelu', 'sigmoid', 'tanh', 'softmax', 'elu', 'gelu', 'swish'
    '''
    act = nn.Identity()
    if activation == 'relu':
        act = nn.ReLU()
    elif activation == 'sigmoid':
        act = nn.Sigmoid()
    elif activation == 'tanh':
        act = nn.Tanh()
    elif activation == 'softmax':
        act = nn.Softmax()
    elif activation == 'elu':
        act = nn.ELU()
    elif activation == 'leakyrelu':
        act = nn.LeakyReLU()
    elif activation == 'gelu':
        act = nn.GELU()
    elif activation == 'silu':
        act = nn.SiLU()
        
    return act


def select_normalization(normalization):
    '''
        This is a wrapped function used for selecting normalization method.
        normalization : a tuple, the first element indicates the normalization type:
        0 : layernorm, the second element should be a tuple corresponding to the normalized shape
        1 : 1d-batchnorm, the second element should be the channels to normalize
        2 : 2d-batchnorm, similarly the second element should be the channels to normalize
        -1 : groupnorm, the second element should be the groups and the third element should be the channels to normalize
                to increase robustness, if channels is not divisible by groups, the groups is modified to the greatest common 
                divisers(gcd).
    '''
    norm = nn.Identity()
    try:
        if normalization[0] == 1:
            norm = nn.BatchNorm1d(normalization[1])
        elif normalization[0] == 2:
            norm = nn.BatchNorm2d(normalization[1])
        elif normalization[0] == 0:
            norm = nn.LayerNorm(normalization[1])    
        elif normalization[0] == -1:
            norm = nn.GroupNorm(num_groups = min(normalization[1], math.gcd(normalization[1], normalization[2])),
                                num_channels = normalization[2])
    except Exception as e:
        pass

    return norm

def get_mask(seq_len):
    '''
        Get a typical GPT attention mask.
    '''
    mask0 = torch.full((seq_len, seq_len), -float('inf'))
    mask1 = torch.triu(mask0, diagonal = 1)
    return mask1

def exists(x):
    '''
        Check whether the input is NoneType.
    '''
    return x is not None

def timer_func(func):
    '''
        A basic timing function.
    '''
    def call(*args, **kwargs):
        st = time.time()
        func(*args, **kwargs)
        ed = time.time()
        dur = ed - st
        print(f"time {dur} seconds costed, which is {dur / 60} minutes or {dur / 3600} hours")
    return call

def cuda_timer_func(func):
    '''
        A basic timing function for cuda.
    '''
    def call(*args, **kwargs):
        start = torch.cuda.Event(enable_timing = True)
        end = torch.cuda.Event(enable_timing = True)
        start.record()
        func(*args, **kwargs)
        end.record()
        dur = end.elapsed_time(start) / 1000
        print(f"time {dur} seconds costed, which is {dur / 60} minutes or {dur / 3600} hours")
    return call

def isnan(x):
    '''
        Check whether a tensor contains a NAN element.
    '''
    return torch.isnan(x).any()

def isinf(x):
    '''
        Check whether a tensor contains a INF element.
    '''
    return torch.isinf(x).any()

def _save_image(img, path, de_norm = True):
    '''
        Save an image to a given path.
        img: the image to save, a tensor with shape (C, H, W) or (1, C, H, W)
        path: the path to save the image
        de_norm: if True, map the elements in img to (0, 1)
    '''
    img = img.squeeze(0).detach().cpu().data
    if de_norm:
        img = ((img + 1) / 2).clamp(0, 1)
    save_image(img, path)

def read_image(path, convert = 'RGB', trans = None):
    '''
        Read an image.
        path : the path to read from
        convert : the type of the image, default RGB
        trans : how to transform this image
    '''
    img = Image.open(path).convert(convert)
    if exists(trans):
        img = trans(img)
    return img

def set_seed(seed = 3407):
    '''
        Set a given random seed.
    '''
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True 


import torch.nn as nn
import torch
from .utils import select_activation, select_normalization, fea2cha, cha2fea

def select_resconnection(in_channels, out_channels):
    mapping = nn.Identity()
    if in_channels != out_channels:
        mapping = nn.Conv2d(in_channels, out_channels, 1, 1, 0)
    return mapping

'''
    This is a typical module in ResNet with 2 conv layers.
'''
class Res_Conv(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, 
                 normalization = None, activation = 'relu', res_connection = True):
        nn.Module.__init__(self)
        self.conv1 = nn.Conv2d(in_channels= in_channels, out_channels= out_channels, 
                            kernel_size= kernel_size, padding= (kernel_size[0] - 1) // 2)
        self.conv2 = nn.Conv2d(in_channels = out_channels, out_channels = out_channels,
                               kernel_size= kernel_size, padding= (kernel_size[0] - 1) // 2)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.res_connection = res_connection
        
        self.norm1 = select_normalization(normalization)
        self.norm2 = select_normalization(normalization)
        self.act = select_activation(activation)
        if self.res_connection:
            self.conv3 = select_resconnection(self.in_channels, self.out_channels)
                
        
    def forward(self, x):
        output = self.conv1(x)
        output = self.norm1(output)
        output = self.act(output)
            
        output = self.conv2(output)
        output = self.norm2(output)
            
        if self.res_connection:
            output = output + self.conv3(x)
            
        return self.act(output)
    
'''
    This is a residual convolution layer with one conv.
'''
class SConv(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, 
                 normalization = None, activation = 'relu', res_connection = True):
        nn.Module.__init__(self)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.res_connection = res_connection
        
        self.conv1 = nn.Conv2d(in_channels= in_channels, out_channels= out_channels, 
                            kernel_size= kernel_size, padding= (kernel_size[0] - 1) // 2)
        if self.res_connection:
            self.conv2 = select_resconnection(self.in_channels, self.out_channels)
        
        self.norm = select_normalization(normalization)
        self.act = select_activation(activation)
                
        
    def forward(self, input):
        output = self.conv1(input)
        output = self.norm(output)
        if self.res_connection:
            output = output + self.conv2(input)
        return self.act(output)
    
'''
    This is a conv module with self-attention layers.
'''
class Att_Conv(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, siz, num_heads, dropout,
                normalization = None, activation = 'relu', res_connection = True):
        nn.Module.__init__(self)
        
        self.conv1 = SConv(in_channels, out_channels, kernel_size, 
                           normalization, activation, res_connection)
        self.conv2 = SConv(out_channels, out_channels, kernel_size, 
                           normalization, activation, res_connection)
        self.norm = select_normalization(normalization)
        self.att = nn.MultiheadAttention(out_channels, num_heads, 
                                         dropout, batch_first = True)
        
        self.in_channels = in_channels
        self.out_channels = out_channels        
        self.res_connection = res_connection
        
        if self.res_connection:
            self.conv3 = select_resconnection(self.in_channels, self.out_channels)
        
        
    def forward(self, x):
        fea = self.conv1(x)
        a_fea = cha2fea(fea)
        att_out, _ = self.att(a_fea, a_fea, a_fea)
        fea = fea2cha(att_out)
        fea = self.norm(fea)
        if self.res_connection:
            fea = fea + self.conv3(x)
        return self.conv2(fea)


import torch.nn as nn
import torch
from .utils import select_activation, select_normalization

'''
    This is a module of a fully connected layer with optional dropout, bias, normalization, 
    activation and residual connection.
'''
class FC(nn.Module):
    def __init__(self, input_dim, output_dim, dropout, bias = True, 
                normalization = None, activation = 'relu', res_connection = False):
        '''
            input_dim: The dimension of the input, which is the size of the last dimension
                        of the input.
            output_dim: The dimension of the output, which is the size of the last dimension
                        of the output.
            dropout: The ratio of dropout. 0 means no dropout. 
            bias: Whether bias is used in Linear Layer.
            normalization: A tuple indicating which kind of normalization is used, 
            3 types are supported: (0, size) indicates layer normalization with size, 
                                    (1, channels) indicates 1d batch normalization on channels,
                                    (2, channels) indicates 2d batch normalization on channels.
            activation: A string indicating which kind of activation function to use. 
                        All letters should be small.
            res_connection: A bool indicating whether residual connection is used, only possible 
                            if input dim equals to output_dim.
        '''
        
        nn.Module.__init__(self)
        
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.res_connection = res_connection
                
        self.li = nn.Linear(input_dim, output_dim, bias = bias)
        self.norm = select_normalization(normalization)
        self.act = select_activation(activation)        
        self.drop = nn.Dropout(dropout)

    def forward(self, X):
        output = self.li(X)
        if self.norm:
            output = self.norm(output)
        if self.input_dim == self.output_dim and self.res_connection:
            output += X
        if self.act:
            output = self.act(output)
        output = self.drop(output)
        return output



import torch
import torch.nn as nn
from torchhandy.Parallel import Parallel_Trainer
import numpy as np
import random
import time
from torchhandy.utils import clear_cuda

class Trainer(object):
    def __init__(self, model, trainset, valset, trainconfig, args, loss_dict):
        self.trainset = trainset
        self.valset = valset
        self.trainer = Parallel_Trainer(args.synch, trainconfig)
        self.seed = trainconfig.seed
        self.resume = args.resume
        self.log_file = self.open_log(trainconfig)
        
        self.resume_path = trainconfig.resume_path
        self.cur_path = trainconfig.cur_path
        self.best_path = trainconfig.best_path
        self.model_name = trainconfig.model_name
        
        self.model = self.trainer.model_parallel(model)
        self.start_epoch = 0
        self.lr = trainconfig.lr
        self.weight_decay = trainconfig.weight_decay
        self.T_max = trainconfig.T_max
        self.eta_min = trainconfig.eta_min
        
        self.epochs = trainconfig.epochs
        self.train_bsz = trainconfig.train_bsz
        self.test_bsz = trainconfig.test_bsz
        self.num_workers = trainconfig.num_workers
        self.accumulation = trainconfig.accumulation
        self.loss_dict = loss_dict
        self.best_val = 1e10
    
        self.set_seed(self.seed)
        
        self.optimizer = torch.optim.AdamW(filter(lambda p : p.requires_grad, self.model.parameters()),
                                    lr = self.lr, weight_decay = self.weight_decay)
        
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(self.optimizer, 
                                                                    T_max = self.T_max, 
                                                                    eta_min = self.eta_min)
        if self.resume:
            self.resume_model()
        
    def open_log(self, trainconfig):
        if self.resume:
            log_file = open(trainconfig.log_path, 'a')
        else:
            log_file = open(trainconfig.log_path, 'w+')
        return log_file
    
    def resume_model(self):
        ckpt = torch.load(self.resume_path, map_location = 'cpu')
        self.model.load_state_dict(ckpt['state_dict'])
        self.start_epoch = ckpt['epoch'] + 1
        self.lr = ckpt['lr']
    
    def set_seed(self, seed):
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.benchmark = False
            torch.backends.cudnn.deterministic = True   
            
    def forward_one_epoch(self, epoch, dataset, type = 'train'):
        if type == 'train':
            self.model.train()
        else:
            self.model.eval()
            
        st_time = time.time()
        self.trainer.call_func(print, f'start {type} model {self.model_name} epoch = {epoch}')
    
        sampler = self.trainer.dataset_parallel(dataset)
        dataloader = self.trainer.get_loader(dataset, epoch, self.train_bsz,
                                            self.num_workers, sampler)
        for k in self.loss_dict:
            self.loss_dict[k] = 0
            
        batches = 0
        
        for data in dataloader:
            batches += 1
            if type == 'train':
                output = self.model(data, self.trainer.get_device())
                losses = self.trainer.get_model(self.model).calc_loss(output, data, self.trainer.get_device())
            else:
                with torch.no_grad():
                    output = self.model(data, self.trainer.get_device())
                    losses = self.trainer.get_model(self.model).calc_loss(output, data, self.trainer.get_device())
            
            for id, k in enumerate(self.loss_dict):
                self.loss_dict[k] += losses[id].item()
                
            if type == 'train':
                batch_loss = losses[-1] / self.accumulation
                batch_loss.backward()
                
                if batches % self.accumulation == 0:
                    self.optimizer.step()
                    self.optimizer.zero_grad()
        
        if type == 'train':
            self.optimizer.step()
            self.optimizer.zero_grad()
            self.scheduler.step()
        
        content = f'{epoch} {type}:\n'
        for k in self.loss_dict:
            content += f'{k} = {self.loss_dict[k] / batches}\n'
        
        self.trainer.write_log(self.log_file, content)
        ckpt = {
            'lr' : self.lr,
            'epoch' : epoch,
            'loss' : self.loss_dict['total_loss'] / batches,
        }
        
        if type == 'train':
            self.trainer.save_model(ckpt, self.model, self.cur_path)
        else:
            if ckpt['loss'] < self.best_val:
                self.best_val = ckpt['loss']
                self.trainer.save_model(ckpt, self.model, self.best_path)
        ed_time = time.time()
        dur = ed_time - st_time
        self.trainer.call_func(print, f'this epoch cost {dur} seconds, which is {dur / 60} minutes or {dur / 3600} hours')
        self.trainer.call_func(self.trainer.get_model(self.model).generate_cases, dataset, 1, 
                            self.trainer.get_device(), f'{self.model_name}_{type}_{epoch}')
        
    def train_model(self):
        for epoch in range(self.start_epoch, self.start_epoch + self.epochs):
            clear_cuda()
            self.forward_one_epoch(epoch, self.trainset, type = 'train')
            self.forward_one_epoch(epoch, self.valset, type = 'valid')
    
    def eval_model(self, dataset):
        if dataset is not None:
            self.forward_one_epoch(0, dataset, type = 'test')
        else:
            self.forward_one_epoch(0, self.valset, type = 'test')
    
        